/*
   MyideDisk - base class for accessing a MyIDE drive

   Copyright (C) 2006 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "MyideDisk.h"
#include "Error.h"

MyideDisk::MyideDisk()
	: super()
{
}

MyideDisk::~MyideDisk()
{
}

bool MyideDisk::ReadBlock256(unsigned int sector, unsigned char* buf)
{
	if (!ReadBlock(sector)) {
		return false;
	}
	unsigned int i;
	for (i=0;i<256;i++) {
		buf[i] = fSectorBuffer[i<<1];
	}
	return true;
}

bool MyideDisk::ReadBlock128(unsigned int sector, unsigned char* buf)
{
	if (!ReadBlock(sector)) {
		return false;
	}
	unsigned int i;
	for (i=0;i<128;i++) {
		buf[i] = fSectorBuffer[i<<1];
	}
	return true;
}

bool MyideDisk::WriteBlock256(unsigned int sector, const unsigned char* buf)
{
	Assert(buf);
	unsigned int i;
	memset(fSectorBuffer, 0, 512);
	for (i=0;i<256;i++) {
		fSectorBuffer[i<<1] = buf[i];
	}
	if (!WriteBlock(sector)) {
		return false;
	}
	return true;
}

bool MyideDisk::WriteBlock512(unsigned int sector, const unsigned char* buf)
{
	Assert(buf);
	memcpy(fSectorBuffer, buf, 512);
	if (!WriteBlock(sector)) {
		return false;
	}
	return true;
}

bool MyideDisk::WriteBlock128(unsigned int sector, const unsigned char* buf)
{
	unsigned int i;
	memset(fSectorBuffer, 0, 512);
	for (i=0;i<128;i++) {
		fSectorBuffer[i<<1] = buf[i];
	}
	if (!WriteBlock(sector)) {
		return false;
	}
	return true;
}


bool MyideDisk::ReadPartitionTable()
{
	unsigned char buf[256];

	if (!ReadBlock256(0, buf)) {
		return false;
	}

	fCylinders = buf[0] + 256 * buf[1] + 1;
	fHeads = buf[2] + 1;
	fSectors = buf[3];
	
	fSectorsPerCylinder = fHeads * fSectors;

	fNumberOfPartitions = buf[4];

	fDriveBits = buf[5];
	fOptionByte = buf[14];

	fImageSpaceCylinder = 256 * buf[6];
	fCylindersPerImageSlot = buf[7];

	for (int i=0;i<eMaxPartitions;i++) {
		int idx = i*8+8;
		fPartitionInfo[i].DriveNumber = buf[idx];
		fPartitionInfo[i].StartingCylinder = buf[idx+1] + 256 * buf[idx+2];
		fPartitionInfo[i].EndingCylinder = buf[idx+3] + 256 * buf[idx+4];
		fPartitionInfo[i].MaxSectors =
			(fPartitionInfo[i].EndingCylinder - fPartitionInfo[i].StartingCylinder) *
			fSectorsPerCylinder;
		if (fPartitionInfo[i].MaxSectors > 65535) {
			fPartitionInfo[i].MaxSectors = 65535;
		}
		if (buf[idx+5] == 1) {
			fPartitionInfo[i].Density = eDD;
		} else {
			fPartitionInfo[i].Density = eSD;
		}
		if (fPartitionInfo[i].DriveNumber >= 1 && fPartitionInfo[i].DriveNumber <=8
			&& (fDriveBits & (1<<(fPartitionInfo[i].DriveNumber - 1)))) {
			fPartitionInfo[i].IsActive = true;
		} else {
			fPartitionInfo[i].IsActive = false;
		}
	}
	return true;
}

void MyideDisk::UpdateDriveBits()
{
	fDriveBits = 0;
	for (int i=0; i<8; i++) {
		if (fPartitionInfo[i].DriveNumber >= 1 && fPartitionInfo[i].DriveNumber <= 8
			&& fPartitionInfo[i].IsActive) {
			fDriveBits |= 1 << (fPartitionInfo[i].DriveNumber-1);
		}
	}
}

bool MyideDisk::WritePartitionTable()
{
	unsigned char buf[256];
	memset(buf, 0, 256);

	buf[0] = (fCylinders - 1) & 0xff;
	buf[1] = (fCylinders - 1) >> 8;
	buf[2] = fHeads - 1;
	buf[3] = fSectors;

	buf[4] = fNumberOfPartitions;
	
	buf[5] = fDriveBits;

	buf[6] = fImageSpaceCylinder >> 8;
	buf[7] = fCylindersPerImageSlot;

	buf[14] = fOptionByte;

	for (int i=0;i<eMaxPartitions;i++) {
		int idx = i*8+8;
		buf[idx] = fPartitionInfo[i].DriveNumber;
		buf[idx+1] = fPartitionInfo[i].StartingCylinder & 0xff;
		buf[idx+2] = fPartitionInfo[i].StartingCylinder >> 8;
		buf[idx+3] = fPartitionInfo[i].EndingCylinder & 0xff;
		buf[idx+4] = fPartitionInfo[i].EndingCylinder >> 8;
		if (fPartitionInfo[i].Density == eDD) {
			buf[idx+5] = 1;
		} else {
			buf[idx+5] = 0;
		}
	}

	return WriteBlock256(0, buf);
}

bool MyideDisk::ReadImageBlock(unsigned int imageno, unsigned int sector)
{
	if (!ImageSpaceEnabled()) {
		return false;
	}
	if (!IsValidImageNumber(imageno)) {
		return false;
	}
	if (!IsValidImageSector(sector)) {
		return false;
	}

	unsigned int disk_secno =
		(fImageSpaceCylinder + (imageno - 1) * fCylindersPerImageSlot) * fSectorsPerCylinder
		+ sector - 1;
	return ReadBlock(disk_secno);
}

bool MyideDisk::WriteImageBlock(unsigned int imageno, unsigned int sector)
{
	if (!ImageSpaceEnabled()) {
		return false;
	}
	if (!IsValidImageNumber(imageno)) {
		return false;
	}
	if (!IsValidImageSector(sector)) {
		return false;
	}

	unsigned int disk_secno =
		(fImageSpaceCylinder + (imageno - 1) * fCylindersPerImageSlot) * fSectorsPerCylinder
		+ sector - 1;
	return WriteBlock(disk_secno);
}

bool MyideDisk::ReadImageSector(unsigned int imageno, unsigned int sector, unsigned char* buf, unsigned int seclen)
{
	if (seclen > 256) {
		return false;
	}
	if (!ReadImageBlock(imageno, sector)) {
		return false;
	}
	for (unsigned int i=0;i<seclen;i++) {
		buf[i] = fSectorBuffer[i<<1];
	}
	return true;
}

bool MyideDisk::WriteImageSector(unsigned int imageno, unsigned int sector, const unsigned char* buf, unsigned int seclen)
{
	if (seclen > 256) {
		return false;
	}
	memset(fSectorBuffer, 0, 512);
	for (unsigned int i=0;i<seclen;i++) {
		fSectorBuffer[i<<1] = buf[i];
	}

	return WriteImageBlock(imageno, sector);
}

bool MyideDisk::ReadPartitionBlock(unsigned int partno, unsigned int sector)
{
	if (!IsValidPartitionNumber(partno)) {
		return false;
	}
	if (!fPartitionInfo[partno-1].IsActive) {
		return false;
	}
	if (!IsValidPartitionSector(partno, sector)) {
		return false;
	}

	unsigned int disk_secno =
		fPartitionInfo[partno-1].StartingCylinder * fSectorsPerCylinder
		+ sector - 1;
	return ReadBlock(disk_secno);
}

bool MyideDisk::WritePartitionBlock(unsigned int partno, unsigned int sector)
{
	if (!IsValidPartitionNumber(partno)) {
		return false;
	}
	if (!fPartitionInfo[partno-1].IsActive) {
		return false;
	}
	if (!IsValidPartitionSector(partno, sector)) {
		return false;
	}

	unsigned int disk_secno =
		fPartitionInfo[partno-1].StartingCylinder * fSectorsPerCylinder
		+ sector - 1;
	return WriteBlock(disk_secno);
}

bool MyideDisk::ReadPartitionSector(unsigned int partno, unsigned int sector, unsigned char* buf, unsigned int seclen)
{
	if (seclen > 256) {
		return false;
	}
	if (!ReadPartitionBlock(partno, sector)) {
		return false;
	}
	for (unsigned int i=0;i<seclen;i++) {
		buf[i] = fSectorBuffer[i<<1];
	}
	return true;
}

bool MyideDisk::WritePartitionSector(unsigned int partno, unsigned int sector, const unsigned char* buf, unsigned int seclen)
{
	if (seclen > 256) {
		return false;
	}
	memset(fSectorBuffer, 0, 512);
	for (unsigned int i=0;i<seclen;i++) {
		fSectorBuffer[i<<1] = buf[i];
	}

	return WritePartitionBlock(partno, sector);
}

bool MyideDisk::ClearImageInfo(unsigned int imageno)
{
	unsigned char buf[256];

	unsigned int sector = fSectorsPerCylinder * fCylindersPerImageSlot;

	if (!ImageSpaceEnabled()) {
		return false;
	}
	if (!IsValidImageNumber(imageno)) {
		return false;
	}
	if (!IsValidImageSector(sector)) {
		return false;
	}
	memset(buf, 0, 256);

	if (!WriteImageSector(imageno, sector, buf, 256)) {
		return false;
	}
	return true;
}


bool MyideDisk::GetImageInfo(unsigned int imageno, MyideImage& img)
{
	unsigned char buf[256];
	if (!ReadImageSector(imageno, fSectorsPerCylinder * fCylindersPerImageSlot, buf, 256)) {
		return false;
	}
	int i=0,j=0;
	char c;
	while (i<29) {
		c = buf[i] & 0x7f;
		i++;
		if (c < 64) {
			c += 32;
		} else if (c < 96) {
			c -= 64;
		}
		if (c < 32 || c == 96 || c == 123 || c >= 125) {
			continue;
		}
		img.Name[j++] = c;
	}
	img.Name[j] = 0;
	j--;
	while (j >= 0 && img.Name[j] == 32) {
		img.Name[j--] = 0;
	}
	switch (buf[30]) {
	case 0xa4:
		img.Density = eDD; break;
	case 0xad:
		img.Density = eED; break;
	case 0xb3:
		img.Density = eSD; break;
	default:
		img.Density = eUnknownDensity; break;
	}
	return true;
}

bool MyideDisk::SetImageInfo(unsigned int imageno, const MyideImage& img)
{
	unsigned char buf[256];
	unsigned int sector = fSectorsPerCylinder * fCylindersPerImageSlot;

	if (!ImageSpaceEnabled()) {
		return false;
	}
	if (!IsValidImageNumber(imageno)) {
		return false;
	}
	if (!IsValidImageSector(sector)) {
		return false;
	}
	if (img.Density == eUnknownDensity) {
		return false;
	}

	memset(buf, 0, 256);

	switch (img.Density) {
	case eSD: buf[30] = 0xb3; break;
	case eED: buf[30] = 0xad; break;
	case eDD: buf[30] = 0xa4; break;
	default: break;
	}

	int i=0,j=0;
	char c;
	while (i < 30 && j < 29 && (c=img.Name[i])) {
		i++;
		if (c < 32 || c == 96 || c == 123 || c >= 125) {
			continue;
		}
		if (c < 32) {
			c += 64;
		} else if (c < 96) {
			c -= 32;
		}
		buf[j++] = c;
	}

	if (!WriteImageSector(imageno, sector, buf, 256)) {
		return false;
	}
	return true;
}

