#ifndef MYIDEDISK_H
#define MYIDEDISK_H

/*
   MyideDisk - base class for accessing a MyIDE drive

   Copyright (C) 2006 Matthias Reichl <hias@horus.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include "RefCounted.h"
#include "AtariDebug.h"

enum EDensity {
	eSD = 0,
	eDD = 1,
	eED = 2,
	eUnknownDensity = -1
};

typedef struct {
	unsigned int DriveNumber;
	unsigned int StartingCylinder;
	unsigned int EndingCylinder;
	unsigned int MaxSectors;
	EDensity Density;
	bool IsActive;
} MyidePartition;

typedef struct {
	EDensity Density;
	char Name[30];
} MyideImage;

class MyideDisk : public RefCounted {
public:
	inline unsigned int GetNumberOfCylinders() const;
	inline unsigned int GetNumberOfHeads() const;
	inline unsigned int GetNumberOfSectors() const;
	inline unsigned int GetNumberOfSectorsPerCylinder() const;

	enum { eMaxPartitions = 8 };

	inline unsigned int GetNumberOfPartitions() const;
	inline bool GetPartitionInfo(unsigned int number, MyidePartition& part) const;
	inline bool SetPartitionInfo(unsigned int number, const MyidePartition& part);

	bool ReadPartitionTable();
	bool WritePartitionTable();

	inline bool ImageSpaceEnabled() const;
	inline unsigned int GetImageSpaceStartingCylinder() const;
	inline unsigned int GetCylindersPerImageSlot() const;
	// number of sectors per image slot, excluding the last sector (for image infos)
	inline unsigned int GetSectorsPerImageSlot() const;
	inline unsigned int GetNumberOfImageSlots() const;

	inline unsigned char GetOptionByte() const;
	inline unsigned char GetDriveBits() const;

	bool ClearImageInfo(unsigned int imageno);
	bool GetImageInfo(unsigned int imageno, MyideImage& img);
	bool SetImageInfo(unsigned int imageno, const MyideImage& img);

	bool ReadImageSector(unsigned int imageno, unsigned int sector, unsigned char* buf, unsigned int seclen);
	bool WriteImageSector(unsigned int imageno, unsigned int sector, const unsigned char* buf, unsigned int seclen);

	bool ReadPartitionSector(unsigned int partno, unsigned int sector, unsigned char* buf, unsigned int seclen);
	bool WritePartitionSector(unsigned int partno, unsigned int sector, const unsigned char* buf, unsigned int seclen);

	inline bool IsValidImageNumber(unsigned int imgno) const;
	inline bool IsValidImageSector(unsigned int secno) const;

	inline bool IsValidPartitionNumber(unsigned int partno) const;
	inline bool IsValidPartitionSector(unsigned int partno, unsigned int secno) const;

	inline bool IsValidSectorLength(unsigned int sector, EDensity density, unsigned int len);

	// direct disk access - be careful!
	bool WriteBlock256(unsigned int sector, const unsigned char* buf);
	bool WriteBlock512(unsigned int sector, const unsigned char* buf);

protected:
	MyideDisk();
	virtual ~MyideDisk();

	virtual bool ReadBlock(unsigned int sector) = 0;
	virtual bool WriteBlock(unsigned int sector) = 0;

	mutable unsigned char fSectorBuffer[512];

private:
	typedef RefCounted super;

	bool ReadImageBlock(unsigned int imgno, unsigned int sector);
	bool WriteImageBlock(unsigned int imgno, unsigned int sector);

	bool ReadPartitionBlock(unsigned int partno, unsigned int sector);
	bool WritePartitionBlock(unsigned int partno, unsigned int sector);

	bool ReadBlock256(unsigned int sector, unsigned char* buf);
	bool ReadBlock128(unsigned int sector, unsigned char* buf);

	bool WriteBlock128(unsigned int sector, const unsigned char* buf);

	void UpdateDriveBits();

	// physical drive settings
	unsigned int fCylinders;
	unsigned int fHeads;
	unsigned int fSectors;

	unsigned int fSectorsPerCylinder;

	unsigned int fNumberOfPartitions;
	MyidePartition fPartitionInfo[eMaxPartitions];

	unsigned char fOptionByte;
	unsigned char fDriveBits;

	unsigned int fImageSpaceCylinder;
	unsigned int fCylindersPerImageSlot;

};

inline unsigned int MyideDisk::GetNumberOfCylinders() const
{
	return fCylinders;
}

inline unsigned int MyideDisk::GetNumberOfHeads() const
{
	return fHeads;
}

inline unsigned int MyideDisk::GetNumberOfSectors() const
{
	return fSectors;
}

inline unsigned int MyideDisk::GetNumberOfSectorsPerCylinder() const
{
	return fSectorsPerCylinder;
}

inline unsigned int MyideDisk::GetNumberOfPartitions() const
{
	return fNumberOfPartitions;
}

inline bool MyideDisk::GetPartitionInfo(unsigned int number, MyidePartition&part) const
{
	if ((number < 1) || (number > eMaxPartitions)) {
		return false;
	} else {
		part = fPartitionInfo[number - 1];
		return true;
	}
}

inline bool MyideDisk::SetPartitionInfo(unsigned int number, const MyidePartition&part)
{
	if ((number < 1) || (number > eMaxPartitions)) {
		return false;
	} else {
		fPartitionInfo[number - 1] = part;
		UpdateDriveBits();
		return true;
	}
}

inline unsigned int MyideDisk::GetImageSpaceStartingCylinder() const
{
	return fImageSpaceCylinder;
}

inline unsigned int MyideDisk::GetCylindersPerImageSlot() const
{
	return fCylindersPerImageSlot;
}

inline unsigned int MyideDisk::GetSectorsPerImageSlot() const
{
	return fCylindersPerImageSlot * fSectorsPerCylinder - 1;
}

inline bool MyideDisk::ImageSpaceEnabled() const
{
	return fImageSpaceCylinder && fCylindersPerImageSlot && (fImageSpaceCylinder < fCylinders);
}

inline unsigned int MyideDisk::GetNumberOfImageSlots() const
{
	if (!ImageSpaceEnabled()) {
		return false;
	}
	return (fCylinders - fImageSpaceCylinder) / fCylindersPerImageSlot;
}

inline bool MyideDisk::IsValidImageNumber(unsigned int imgno) const
{
	return (imgno >= 1) && (imgno <= GetNumberOfImageSlots());
}

inline bool MyideDisk::IsValidImageSector(unsigned int secno) const
{
	return (secno >= 1) && (secno <= GetNumberOfSectorsPerCylinder() * GetCylindersPerImageSlot());
}

inline bool MyideDisk::IsValidPartitionNumber(unsigned int partno) const
{
	return (partno >= 1) && (partno <= eMaxPartitions);
}

inline bool MyideDisk::IsValidPartitionSector(unsigned int partno, unsigned int secno) const
{
	if (!IsValidPartitionNumber(partno)) {
		Assert(false);
		return false;
	}
	return (secno >= 1) && (secno <= fPartitionInfo[partno-1].MaxSectors);
}

inline bool MyideDisk::IsValidSectorLength(unsigned int sector, EDensity density, unsigned int seclen)
{
        if (density == eDD) {
		if (sector <= 3 && seclen != 128) {
			return false;
		}
		if (sector > 3 && seclen != 256) {
			return false;
		}
	} else {
		if (seclen != 128) {
			return false;
		}
	}
	return true;
}

inline unsigned char MyideDisk::GetOptionByte() const
{
	return fOptionByte;
}

inline unsigned char MyideDisk::GetDriveBits() const
{
	return fDriveBits;
}

#endif
