#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "MyideDisk.h"
#include "MyOpts.h"

#include "AtrUtils.h"
#include "RCPtr.h"
#include "Error.h"
#include "AtariDebug.h"
#include "ImageManager.h"

#include "SimpleProgressIndicator.h"

#ifdef WIN32
#include <windows.h>
#include <winioctl.h>
#include <winnt.h>
#include "MyideDiskWin32.h"
#else
#include "MyideDiskUnix.h"
#endif

using namespace AtrUtils;

void print_bin(unsigned char c)
{	
	int i;
	for (i=7;i>=0;i--) {
		if (c & (1<<i)) {
			putchar('1');
		} else {
			putchar('0');
		}
	}
}

long get_file_size(FILE *f)
{
	long current_pos = ftell(f);
	fseek(f, 0L, SEEK_END);
	long end_pos = ftell(f);
	fseek(f, current_pos, SEEK_SET);
	return end_pos;
}

void print_partition_table(RCPtr<MyideDisk>& disk)
{
	printf("\nMyIDE partition table:\n");
	printf("%d cylinders %d heads %d sectors (%d sectors per cyl)\n",
		disk->GetNumberOfCylinders(),
		disk->GetNumberOfHeads(),
		disk->GetNumberOfSectors(),
		disk->GetNumberOfHeads() * disk->GetNumberOfSectors());
	printf("Image space: ");
	if (disk->ImageSpaceEnabled()) {
		printf("first cylinder: %d  cylinders per slot: %d  total slots: %d\n",
			disk->GetImageSpaceStartingCylinder(),
			disk->GetCylindersPerImageSlot(),
			disk->GetNumberOfImageSlots());
	} else {
		printf("disabled\n");
	}
	printf("partitions: %d", disk->GetNumberOfPartitions());
	printf(" option byte: "); print_bin(disk->GetOptionByte());
	printf(" drive bits: "); print_bin(disk->GetDriveBits());
	printf("\n");

	printf("part active drv start   end dns sectors\n");
	for (int i=1;i<=MyideDisk::eMaxPartitions;i++) {
		MyidePartition part;
		if (!(disk->GetPartitionInfo(i, part))) {
			printf("error getting partition info!\n");
			continue;
		}
		printf("%d    %s    D%d: %5d %5d %s    %5d\n", i,
			part.IsActive ? "yes" : "no ",
			part.DriveNumber, part.StartingCylinder, part.EndingCylinder,
			part.Density == eDD ? "DD" : "SD",
			part.MaxSectors);
	}
}

void print_single_image_info(const MyideImage& img, unsigned int imgno)
{
	if (img.Name[0] && (img.Density != eUnknownDensity)) {
		printf("%5d: ", imgno);
		switch (img.Density) {
		case eSD: printf("SD"); break;
		case eED: printf("ED"); break;
		case eDD: printf("DD"); break;
		case eUnknownDensity: printf("??"); break;
		default: printf("DD"); break;
		}
		printf(" %s", img.Name);
	}
}

void print_image_infos(RCPtr<MyideDisk>& disk)
{
	unsigned int imgno;
	unsigned int maximg = disk->GetNumberOfImageSlots();

	printf("\nImage slot information:\n");
	printf(" slot# dn name\n");
	if (!disk->ImageSpaceEnabled()) {
		printf("image space is disabled\n");
	} else {
		MyideImage img;
		for (imgno = 1; imgno <= maximg; imgno++) {
			if (disk->GetImageInfo(imgno, img)) {
				if (img.Name[0] && (img.Density != eUnknownDensity)) {
					print_single_image_info(img, imgno);
					printf("\n");
				}
			} else {
				printf("error reading info of image %d\n", imgno);
			}
		}
		printf("\n");
	}
}

bool print_imagemanager_infos(RCPtr<ImageManager>& manager, RCPtr<MyideDisk>& disk)
{
	unsigned int imgno;
	unsigned int maximg = manager->GetNumberOfEntries();

	printf("\nImage Manager information:\n");
	printf("   cat  stat  slot# dn name\n");

	unsigned int i;
	const char* cat;
	MyideImage img;
	for (i=1; i<=maximg; i++) {
		imgno = manager->GetSlotNumberOfEntry(i);
		if (imgno == 0) {
			return false;
		}
		cat = manager->GetCategoryOfEntry(i);
		if (cat == 0) {
			return false;
		}
		if (disk->GetImageInfo(imgno, img)) {
			printf("[%s] [", cat);

			unsigned char status;
			status = manager->GetStatusOfEntry(i);

			switch (status & 0x60) {
			case 0x00: printf("S"); break;
			case ImageManager::eBit1040SectorsDisk: printf("E"); break;
			case ImageManager::eBitDoubleDensity: printf("D"); break;
			default: printf("?"); break;
			}

			if (status & ImageManager::eBitEnableBasic) {
				printf("B");
			} else {
				printf(" ");
			}
			if (status & ImageManager::eBitDeleted) {
				printf("X");
			} else {
				printf(" ");
			}
			printf("] ");

			print_single_image_info(img, imgno);
			printf("\n");
		} else {
			printf("error reading info of image %d\n", imgno);
			return false;
		}
	}
	printf("\n");
	return true;
}

bool add_image_information_to_image_manager(RCPtr<ImageManager>& imgManager, RCPtr<MyideDisk>& disk)
{
	printf("scanning for existing images and adding them to Image Manager\n");
	if (!disk->ImageSpaceEnabled()) {
		printf("image space is disabled\n");
		return false;
	}
	unsigned int imgno = 1;
	unsigned int maximg = disk->GetNumberOfImageSlots();
	MyideImage img;
	bool done = false;

	while (!done && imgno <= maximg) {
		if (disk->GetImageInfo(imgno, img)) {
			if (img.Name[0] && (img.Density != eUnknownDensity)) {
				print_single_image_info(img, imgno);
				printf(": ");
				fflush(stdout);

				if (!imgManager->SetImageInformation(imgno, img.Density, 1)) {
					return false;
				} else {
					printf("OK\n");
				}
			} else {
				done = true;
			}
		} else {
			printf("error reading info of image %d\n", imgno);
			return false;
		}
		imgno++;
	}
	return true;
}

bool write_image_to_file(RCPtr<MyideDisk>& disk,
	unsigned int imageno, EDensity density, const char* filename)
{
	Assert(filename);

	MyideImage img;
	if (!disk->GetImageInfo(imageno, img)) {
		printf("cannot get image info\n");
		return false;
	}

	if (density == eUnknownDensity) {
		density = img.Density;
	}

	unsigned int sectors = 720;
	bool dd = false;
	RCPtr<SimpleProgressIndicator> progress;

	FILE *f = fopen(filename,"wb");
	if (!f) {
		printf("cannot create %s\n", filename);
		return false;
	}
	try {
		switch (density) {
		case eSD: WriteAtrHeader(f, 720, false); break;
		case eED: sectors = 1040; WriteAtrHeader(f, 1040, false); break;
		case eDD: dd = true; WriteAtrHeader(f, 720, true); break;
		default:
			  printf("invalid density\n");
			  goto exit_err;
		}
	}
	catch (ErrorObject& err) {
		fprintf(stderr, "%s\n", err.AsString());
		goto exit_err;
	}

	progress = new SimpleProgressIndicator(sectors);
	progress->DisplayProgress(0);

	unsigned int i;
	unsigned char buf[256];
	unsigned int len;
	for (i=1;i<=sectors;i++) {
		if (dd && (i>3)) {
			len = 256;
		} else {
			len = 128;
		}
		if (!disk->ReadImageSector(imageno, i, buf, len)) {
			printf(" error reading image sector!\n");
			goto exit_err;
		}
		if (fwrite(buf, 1, len, f) != len) {
			printf(" error writing ATR file\n");
			goto exit_err;
		}
		progress->DisplayProgress(i);
	}
	progress->FinishProgress();
	printf(" OK\n");

	fclose(f);
	return true;

exit_err:
	fclose(f);
	return false;
}

bool write_myide_to_file(RCPtr<MyideDisk>& disk,
	bool isPartition, unsigned int number,
	EDensity density, unsigned int sectors, const char* filename)
{
	Assert(filename);
	Assert(sectors > 0 && sectors <= 65535);
	Assert(density == eSD || density == eDD);

	printf("writing %s %d to %s: ",
		isPartition ? "partition" : "image",
		number, filename);

	FILE *f = fopen(filename,"wb");
	RCPtr<SimpleProgressIndicator> progress;
	if (!f) {
		printf("cannot create %s\n", filename);
		return false;
	}
	try {
		WriteAtrHeader(f, sectors, density == eDD);
	}
	catch (ErrorObject& err) {
		fprintf(stderr, "%s\n", err.AsString());
		goto exit_err;
	}

	progress = new SimpleProgressIndicator(sectors);
	progress->DisplayProgress(0);

	unsigned int i;
	unsigned char buf[256];
	unsigned int len;
	bool ok;
	for (i=1;i<=sectors;i++) {
		if ((density == eDD) && (i>3)) {
			len = 256;
		} else {
			len = 128;
		}
		if (isPartition) {
			ok = disk->ReadPartitionSector(number, i, buf, len);
		} else {
			ok = disk->ReadImageSector(number, i, buf, len);
		}
		if (!ok) {
				printf(" error reading image sector!");
				goto exit_err;
		}
		if (fwrite(buf, 1, len, f) != len) {
			printf(" error writing ATR file\n");
			goto exit_err;
		}
		progress->DisplayProgress(i);
	}
	progress->FinishProgress();
	printf(" OK\n");

	fclose(f);
	return true;

exit_err:
	fclose(f);
	return false;
}


bool write_partition_to_file(RCPtr<MyideDisk>& disk,
	unsigned int partno, const char* filename,
	unsigned int maxSectors = 0)
{
	Assert(filename);

	MyidePartition part;
	if (!disk->GetPartitionInfo(partno, part)) {
		printf("cannot get partition info\n");
		return false;
	}

	EDensity density = part.Density;
	unsigned int sectors = part.MaxSectors;

	if (maxSectors && maxSectors < sectors) {
		sectors = maxSectors;
	}

	return write_myide_to_file(disk, true, partno,
		density, sectors, filename);
}

bool write_image_to_file(RCPtr<MyideDisk>& disk,
	unsigned int imageno, const char* filename)
{
	char fbuf[40];

	MyideImage img;
	if (!disk->GetImageInfo(imageno, img)) {
		printf("cannot get image info\n");
		return false;
	}

	EDensity density = img.Density;

	unsigned int sectors = 720;

	if (!filename) {
		if (!img.Name[0]) {
			printf("no filename specified - skipping image %d\n", imageno);
			return false;
		}
		sprintf(fbuf,"%s.atr", img.Name);
		filename = fbuf;
	}

	switch (density) {
	case eSD: break;
	case eDD: break;
	case eED: sectors = 1040; density = eSD; break;
	default:
		  printf("cannot determine density of image %d - skipping\n", imageno);
		  return false;
	}

	return write_myide_to_file(disk, false, imageno,
		density, sectors, filename);
}

bool write_file_to_image(
	RCPtr<MyideDisk>& disk,
	unsigned int imageno, const char* filename,
	RCPtr<ImageManager>& manager,
	unsigned int managerCategory,
	bool managerBasic,
	bool force_non_standard = false)
{
	printf("writing %s to image %d: ",
		filename, imageno);

#ifdef WIN32
	const char* fn = strrchr(filename, '\\');
#else
	const char* fn = strrchr(filename, '/');
#endif

	if (!fn) {
		fn = filename;
	} else {
		fn++;
	}

	MyideImage img;
	unsigned int len = strlen(fn);
	if (len > 29) len = 29;
	memcpy(img.Name, fn, len+1);

	if (len >=4 && strncasecmp(fn+len-4,".atr",4) == 0) {
		img.Name[len-4] = 0;
	}

	FILE* f = fopen(filename, "rb");
	if (!f) {
		printf("cannot open %s\n", filename);
		return false;
	}

	bool isDD;
	unsigned int sectors;
	unsigned int writesectors;
	RCPtr<SimpleProgressIndicator> progress;

	try {
		ReadAtrHeader(f, sectors, isDD);
	}
	catch (ErrorObject& err) {
		fprintf(stderr, "%s\n", err.AsString());
		goto exit_err;
	}

	if (isDD) {
		if (sectors <= 720) {
			img.Density = eDD;
			writesectors = 720;
		} else {
			if (force_non_standard) {
				printf("WARNING: forcing write of %d sectors DD image\n", sectors);
				img.Density=eDD;
				writesectors = sectors;
			} else {
				printf("unsupported disk format\n");
				goto exit_err;
			}
		}
	} else {
		if (sectors <= 720) {
			img.Density = eSD;
			writesectors = 720;
		} else {
			if (sectors <= 1040) {
				img.Density = eED;
				writesectors = 1040;
			} else {
				if (force_non_standard) {
					printf("WARNING: forcing write of %d sectors DD image\n", sectors);
					img.Density=eED;
					writesectors = sectors;
				} else {
					printf("unsupported disk format\n");
					goto exit_err;
				}
			}
		}
	}

	if (writesectors > disk->GetSectorsPerImageSlot()) {
		printf("ATR file too big for image slot\n");
		goto exit_err;
	}

	if (!disk->SetImageInfo(imageno, img)) {
		printf("cannot set image info\n");
		goto exit_err;
	}

	progress = new SimpleProgressIndicator(writesectors);
	progress->DisplayProgress(0);

	unsigned int sec;
	unsigned char buf[256];
	for (sec = 1; sec <= writesectors; sec++) {
		unsigned int len = 128;
		if (isDD && sec > 3) {
			len = 256;
		}
		if (sec <= sectors) {
			if (fread(buf, 1, len, f) != len) {
				printf(" error reading ATR file\n");
				goto exit_err;
			}
		} else {
			memset(buf, 0, len);
		}
		if (!disk->WriteImageSector(imageno, sec, buf, len)) {
			printf(" error writing to image space\n");
			goto exit_err;
		}
		progress->DisplayProgress(sec);
	}
	progress->FinishProgress();
	printf(" OK\n");

	if (manager.IsNotNull()) {
		if (!manager->SetImageInformation(imageno, img.Density, managerBasic, managerCategory)) {
			goto exit_err;
		}
	}

	fclose(f);
	return true;

exit_err:
	fclose(f);
	return false;
}

bool write_file_to_partition(RCPtr<MyideDisk>& disk,
	unsigned int partno, const char* filename)
{
	printf("writing %s to partition %d: ",
		filename, partno);

	MyidePartition part;

	if (!disk->GetPartitionInfo(partno, part)) {
		Assert(false);
		return false;
	}

	FILE* f = fopen(filename, "rb");
	if (!f) {
		printf("cannot open %s\n", filename);
		return false;
	}

	bool isDD;
	EDensity density;
	unsigned int sectors;
	RCPtr<SimpleProgressIndicator> progress;

	try {
		ReadAtrHeader(f, sectors, isDD);
	}
	catch (ErrorObject& err) {
		fprintf(stderr, "%s\n", err.AsString());
		goto exit_err;
	}

	if (isDD) {
		density = eDD;
	} else {
		density = eSD;
	}

	if (density != part.Density) {
		printf("ATR file density doesn't match partition density\n");
		goto exit_err;
	}

	if (sectors > part.MaxSectors) {
		printf("ATR file too big for partition\n");
		goto exit_err;
	}

	progress = new SimpleProgressIndicator(sectors);
	progress->DisplayProgress(0);

	unsigned int sec;
	unsigned char buf[256];
	for (sec = 1; sec <= sectors; sec++) {
		unsigned int len = 128;
		if (isDD && sec > 3) {
			len = 256;
		}
		if (fread(buf, 1, len, f) != len) {
			printf(" error reading ATR file\n");
			goto exit_err;
		}
		if (!disk->WritePartitionSector(partno, sec, buf, len)) {
			printf(" error writing to image space\n");
			goto exit_err;
		}
		progress->DisplayProgress(sec);
	}
	progress->FinishProgress();
	printf(" OK\n");

	fclose(f);
	return true;

exit_err:
	fclose(f);
	return false;
}

bool write_all_partitions(RCPtr<MyideDisk>& disk, unsigned int partno, unsigned int maxsec)
{
	unsigned int p;
	char name[20];
	MyidePartition part;

	for (p = partno; p <= MyideDisk::eMaxPartitions; p++) {
		sprintf(name,"partition%d.atr",p);
		if (!disk->GetPartitionInfo(p, part)) {
			printf("cannot get info of partition %d\n",p);
			return false;
		}
		if (part.MaxSectors == 0
			|| part.StartingCylinder >= disk->GetNumberOfCylinders()
			|| part.EndingCylinder >= disk->GetNumberOfCylinders()) {
			continue;
		}
		if (!write_partition_to_file(disk, p, name, maxsec)) {
			return false;
		}
	}
	return true;
}

bool write_all_images(RCPtr<MyideDisk>& disk, unsigned int imgno)
{
	unsigned int i;
	MyideImage img;

	for (i = imgno; i <= disk->GetNumberOfImageSlots(); i++) {
		if (!disk->GetImageInfo(i, img)) {
			printf("cannot get info of image %d\n",i);
			return false;
		}
		if (img.Name[0] == 0) {
			continue;
		}
		if (img.Density == eUnknownDensity) {
			continue;
		}
		if (!write_image_to_file(disk, i, NULL)) {
			return false;
		}
	}
	return true;
}

bool write_movie(RCPtr<MyideDisk>& disk, unsigned int cylinder, const char* video_file, const char* audio_file, bool mode_cf=false,
		unsigned int audio_blocks = 3, unsigned int video_blocks = 27)
{
	FILE* fv = NULL;
	FILE* fa = NULL;

	unsigned int secPerCyl = disk->GetNumberOfSectorsPerCylinder();
	unsigned int sector = cylinder * secPerCyl;

	Assert((video_file && video_blocks) || (audio_file && audio_blocks));

	if (mode_cf && ((audio_blocks + video_blocks) & 1)) {
		printf("frame size must be a multiple of 2 in compact flash mode!\n");
		return false;
	}

	if (video_file && video_blocks) {
		if (!(fv = fopen(video_file,"rb"))) {
			printf("cannot open video_file file %s\n", video_file);
			return false;
		}
	}

	if (audio_file && audio_blocks) {
		if (!(fa = fopen(audio_file,"rb"))) {
			printf("cannot open audio file %s\n", audio_file);
			fclose(fv);
			return false;
		}
	}

	printf("writing video_file to cylinder %d", cylinder);
	if (mode_cf) {
		printf(" in compact flash mode");
	}
	printf("\n%d audio blocks %d video blocks per frame\n", audio_blocks, video_blocks);
	if (video_file) {
		printf("video: %s\n", video_file);
	}
	if (audio_file) {
		printf("audio: %s\n", audio_file);
	}

	unsigned char *buf = new unsigned char[(audio_blocks + video_blocks)*256];

	unsigned int i;
	bool ok = true;
	unsigned int frames = 0;
	unsigned int count;

	long est_frames = 0;
	if (video_file) {
		est_frames += get_file_size(fv);
	}
	if (audio_file) {
		est_frames += get_file_size(fa);
	}
	est_frames /= ((long)(audio_blocks + video_blocks)) * 256;

	RCPtr<SimpleProgressIndicator> progress = new SimpleProgressIndicator(est_frames);

	count = audio_blocks + video_blocks;
	if (mode_cf) {
		count >>= 1;
	}

	progress->DisplayProgress(0);
	while (ok) {
		if (fa) {
			if (fread(buf, 256, audio_blocks, fa) != audio_blocks) {
				break;
			}
		}
		if (fv) {
			if (fread(buf + 256*audio_blocks, 256, video_blocks, fv) != video_blocks) {
				break;
			}
		}

		for (i=0;i<count;i++) {
			if (mode_cf) {
				ok = disk->WriteBlock512(sector, buf + i*512);
			} else {
				ok = disk->WriteBlock256(sector, buf + i*256);
			}
			if (!ok) {
				printf(" error writing to image space");
				break;
			}
			sector++;
		}

		if (!ok) {
			break;
		}
		frames++;
		progress->DisplayProgress(frames);
	}
	progress->FinishProgress();
	if (ok) {
		printf(" OK\n");
	}
	if (frames) {
		unsigned int wrote_secs = frames * (audio_blocks + video_blocks);
		if (mode_cf) {
			wrote_secs >>= 1;
		}
		unsigned int num_cyls = (wrote_secs + secPerCyl-1) / secPerCyl;
		printf("wrote a total of %d frames (%d cylinders: %d - %d)\n",
			frames, num_cyls, cylinder, cylinder + num_cyls - 1);
	} else {
		printf("wrote 0 frames - please check your settings!\n");
	}
		
	if (fa) {
		fclose(fa);
	}
	if (fv) {
		fclose(fv);
	}
	delete[] buf;
	return ok;
}

bool activate_partition(RCPtr<MyideDisk>& disk, unsigned int partno, bool active)
{
	if (!disk->IsValidPartitionNumber(partno)) {
		printf("invalid partition number %d\n", partno);
		return false;
	}
	MyidePartition part;
	if (!disk->GetPartitionInfo(partno, part)) {
		printf("getting info for partition %d failed\n", partno);
		return false;
	}
	part.IsActive = active;
	if (!disk->SetPartitionInfo(partno, part)) {
		printf("setting info for partition %d failed\n", partno);
		return false;
	}
	return false;
}

#ifdef WIN32
int list_physical_drives()
{
	int driveno = 0;
	char filename[100];
	HANDLE h;
	DISK_GEOMETRY geometry;
	DWORD length;

	while (true) {
		sprintf(filename, "\\\\.\\PhysicalDrive%d",driveno);
		h = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, 0);
		if (h == INVALID_HANDLE_VALUE) {
			int err = GetLastError();
			if (err == ERROR_ACCESS_DENIED) {
				printf("error - need admin privileges to access phyiscal drives!\n");
				return 1;
			}
			if (err == ERROR_FILE_NOT_FOUND) {
				break;
			}
			printf("error opening %s\n", filename);
			return 1;
		} else {
			printf("%s: ", filename);
			if (!DeviceIoControl(h, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0,
				&geometry, sizeof(geometry), &length, NULL)) {
				printf("reading drive geometry failed\n");
			} else {
				printf("%ld/%ld/%ld C/H/S, %ldMB\n",
					(long) geometry.Cylinders.QuadPart,
					geometry.TracksPerCylinder,
					geometry.SectorsPerTrack,
					(long)((geometry.Cylinders.QuadPart * geometry.TracksPerCylinder * geometry.SectorsPerTrack 
					* geometry.BytesPerSector) / (1024*1024)));
			}
			CloseHandle(h);
		}
		driveno++;
	}
	return 0;
}
#endif

int main(int argc, char** argv)
{
	RCPtr<MyideDisk> disk;

	enum EMode {
		eNoMode = 0,
		ePartitionMode = 1,
		eImageMode = 2
	};

	EMode mode = eNoMode;
	bool writeMode = false;
	unsigned int audio_blocks = 3;
	unsigned int video_blocks = 27;

	RCPtr<ImageManager> imgManager;

	bool force_write_nonstandard_image = false;

	bool got_number = false;
	unsigned int number;
	unsigned int managerCategory = 0;
	bool managerBasic = false;
	unsigned int maxSector = 0;
	bool ok = true;
	printf("MyIDE Tool V0.30\n(c) 2006-2007 by Matthias Reichl <hias@horus.com>\n");

	const char* arg;
	RCPtr<MyOpts> myopts = new MyOpts(argc, argv, true);

	if (!(arg = myopts->GetOpt())) {
		goto usage;
	}

#ifdef WIN32
	if (strcmp(arg,"--list") == 0) {
		return list_physical_drives();
	}
#endif

	try {
#ifdef WIN32
		disk = new MyideDiskWin32(arg);
#else
		disk = new MyideDiskUnix(arg);
#endif
	}
	catch (ErrorObject& err) {
		fprintf(stderr, "%s\n", err.AsString());
		exit(1);
	}

	while ((arg = myopts->GetOpt())) {
		// image manager modes
		if (strncmp(arg, "-I", 2) == 0) {
			if (imgManager.IsNull()) {
				imgManager = new ImageManager();
				switch (arg[2]) {
				case 0:
				case 'L':
				case 'c':
				case 'C':
				case 's':
				case 'b':
				case 'B':
					if (!imgManager->ReadFromDisk(disk)) {
						printf("couldn't read Image Manager information\n");
						return 1;
					}
					break;
				default:
					break;
				}
			}
			switch (arg[2]) {
			case 'b':
				managerBasic = true;
				break;
			case 'B':
				managerBasic = false;
				break;
			case 'c':
				if ((arg = myopts->GetOpt())) {
					managerCategory = strtol(arg, NULL, 0);
					if (managerCategory < 1 || managerCategory > ImageManager::eNumCategories) {
						printf("illegal Image Manager category %d\n", managerCategory);
						return 1;
					}
				} else {
					goto usage;
				}
				break;
			case 'C':
				{
					printf("Image Manager categories:\n");
					int i;
					for (i=1; i<=ImageManager::eNumCategories; i++) {
						printf("%2d : %s\n", i, imgManager->GetCategoryName(i));
					}
					printf("\n");
				}
				break;
			case 'i':
				imgManager->InitBlank();
				if (!add_image_information_to_image_manager(imgManager, disk)) {
					return 1;
				}
				break;
			case 'I':
				imgManager->InitBlank();
				break;
			case 'L':
				if (!print_imagemanager_infos(imgManager, disk)) {
					return 1;
				}
				break;
			case 's':
				switch(arg[3]) {
				case 'c':
					imgManager->SortImageTable(disk, ImageManager::eSortCategoryAscending);
					break;
				case 'C':
					imgManager->SortImageTable(disk, ImageManager::eSortCategoryDescending);
					break;
				case 0:
				case 'n':
					imgManager->SortImageTable(disk, ImageManager::eSortNameAscending);
					break;
				case 'N':
					imgManager->SortImageTable(disk, ImageManager::eSortNameDescending);
					break;
				case 's':
					imgManager->SortImageTable(disk, ImageManager::eSortSlotNumber);
					break;
				default:
					goto usage;
				}
				break;
			case 0:
				break;
			default:
				printf("illegal option %s\n", arg);
				goto usage;
			}
			mode = eImageMode;
			continue;
		}
		if (strcmp(arg,"-w") == 0) {
			writeMode = true;
			continue;
		}
		if (strcmp(arg,"-b") == 0) {
			writeMode = false;
			continue;
		}
		if (strcmp(arg,"-l") == 0) {
			print_partition_table(disk);
			continue;
		}
		if (strcmp(arg,"-L") == 0) {
			print_image_infos(disk);
			continue;
		}
		if (strcmp(arg,"-p") == 0) {
			if ((arg = myopts->GetOpt())) {
				number = strtol(arg, NULL, 0);
				if (number < 1 || number > MyideDisk::eMaxPartitions) {
					printf("illegal partition number %d\n", number);
					return 1;
				}
			} else {
				goto usage;
			}
			mode = ePartitionMode;
			got_number = true;
			continue;
		}
		if (strcmp(arg,"-a") == 0) {
			if (mode != ePartitionMode) {
				printf("please select a partition before using option '-a'\n");
				goto usage;
			}
			if ((arg = myopts->GetOpt())) {
				switch (*arg) {
				case '0': ok = activate_partition(disk, number, false); break;
				case '1': ok = activate_partition(disk, number, true); break;
				default: goto usage;
				}
			}
			continue;
		}
		if (strcmp(arg,"-W") == 0) {
			ok = disk->WritePartitionTable();
			continue;
		}
		if (strcmp(arg,"-i") == 0) {
			if (!disk->ImageSpaceEnabled()) {
				printf("cannot set image mode, image space is disabled\n");
				return 1;
			} else {
				if ((arg = myopts->GetOpt())) {
					number = strtol(arg, NULL, 0);
					if (number < 1 || number > disk->GetNumberOfImageSlots()) {
						printf("illegal image slot number %d\n", number);
						return 1;
					}
				} else {
					goto usage;
				}
			}
			mode = eImageMode;
			got_number = true;
			continue;
		}
		if (strcmp(arg,"-e") == 0) {
			if (!disk->ImageSpaceEnabled()) {
				printf("cannot set image mode, image space is disabled\n");
				return 1;
			} else {
				if ((arg = myopts->GetOpt())) {
					char* tmpstr = strdup(arg);
					char* dash = strchr(tmpstr,'-');
					unsigned int start, end;
					if (dash) {
						*dash = 0;
						start = strtol(tmpstr, NULL, 0);
						end = strtol(dash+1, NULL, 0);
					} else {
						start = strtol(tmpstr, NULL, 0);
						end = start;
					}
					free(tmpstr);
					if (end < start) {
						printf("end slot lower than start slot\n");
						return 1;
					}
					if (!disk->IsValidImageNumber(start)) {
						printf("invalid starting slot %d\n", start);
						return 1;
					}
					if (!disk->IsValidImageNumber(end)) {
						printf("invalid ending slot %d\n", start);
						return 1;
					}
					unsigned int i;
					for (i=start; i<=end; i++) {
						printf("erasing image info of slot %d: ", i);
						fflush(stdout);
						if (disk->ClearImageInfo(i)) {
							printf("OK\n");
						} else {
							printf("ERROR\n");
							return 1;
						}
					}
				} else {
					goto usage;
				}
			}
			mode = eImageMode;
			got_number = true;
			continue;
		}
		if (strcmp(arg,"-s") == 0) {
			if ((arg = myopts->GetOpt())) {
				maxSector = strtol(arg, NULL, 0);
				if (maxSector > 65535) {
					printf("illegal maximum sector number %d\n", maxSector);
					return 1;
				}
			} else {
				goto usage;
			}
			continue;
		}
		if (strcmp(arg,"-B") == 0) {
			switch (mode) {
			case ePartitionMode:
				ok = write_all_partitions(disk, number, maxSector);
				break;
			case eImageMode:
				ok = write_all_images(disk, number);
				break;
			default:
				goto error_mode;
			}
			if (!ok) return 1;
			continue;
		}
		if (strcmp(arg,"-fi") == 0) {
			force_write_nonstandard_image = true;
			continue;
		}
		if (strncmp(arg,"-m",2) == 0) {
			switch (arg[2]) {
			case 'a':
				if (!(arg = myopts->GetOpt())) {
					goto usage;
				}
				audio_blocks = strtol(arg, NULL, 0);
				continue;
			case 'v':
				if (!(arg = myopts->GetOpt())) {
					goto usage;
				}
				video_blocks = strtol(arg, NULL, 0);
				continue;
			case 'c':
			case 0: {
				char* fileMov = NULL;
				char* fileAudio = NULL;
				unsigned int cylinder;
				bool mode_cf = arg[2] == 'c';
				if (audio_blocks + video_blocks == 0) {
					printf("audio blocks and video blocks cannot both be 0\n");
					return 1;
				}
				if (!(arg = myopts->GetOpt())) {
					goto usage;
				}
				cylinder = strtol(arg, NULL, 0);
				if (video_blocks) {
					if (!(arg = myopts->GetOpt())) {
						goto usage;
					}
					fileMov = strdup(arg);
				}
				if (audio_blocks) {
					if (!(arg = myopts->GetOpt())) {
						if (fileMov) {
							free(fileMov);
						}
						goto usage;
					}
					fileAudio = strdup(arg);
				}
				ok = write_movie(disk, cylinder, fileMov, fileAudio, mode_cf, audio_blocks, video_blocks);
				if (fileMov) {
					free(fileMov);
				}
				if (fileAudio) {
					free(fileAudio);
				}
				if (!ok) return 1;
				continue;
			}
			default:
				goto usage;
			}
		}
		if (*arg == '-') {
			printf("illegal option %s\n", arg);
			goto usage;
		}
		switch (mode) {
		case ePartitionMode:
			if (writeMode) {
				ok = write_file_to_partition(disk, number, arg);
			} else {
				ok = write_partition_to_file(disk, number, arg, maxSector);
			}
			break;
		case eImageMode:
			if (writeMode) {
				if (imgManager.IsNotNull()) {
					if (!got_number) {
						number = imgManager->GetNumberOfEntries() + 1;
						got_number = true;
					}
					if (number > imgManager->GetNumberOfEntries() + 1) {
						printf("non-consecutive images are not allowed in Image Manager mode\n");
						return 1;
					}
				}
				ok = write_file_to_image(disk, number, arg, imgManager, managerCategory, managerBasic, force_write_nonstandard_image);
			} else {
				ok = write_image_to_file(disk, number, arg);
			}
			break;
		default:
			goto error_mode;
		}
		if (!ok) return 1;
		number++;
	}
	if (imgManager.IsNotNull() && imgManager->IsChanged()) {
		printf("writing Image Manager information: ");
		fflush(stdout);
		if (imgManager->WriteToDisk(disk)) {
			printf("OK\n");
		}
	}
	return 0;

error_mode:
	printf("\nyou must specify either image or partition mode\n");
	return 1;

usage:
#ifdef WIN32
	printf("\nusage: myidetool --list | device options,files...\n");
#else
	printf("\nusage: myidetool device options,files...\n");
#endif
	printf("currently supported options are:\n");
#ifdef WIN32
	printf("  --list     list all available harddrives\n");
#endif
	printf("  -l         list partition table\n");
	printf("  -L         list image slots\n");
	printf("  -i number  select image slot <number>\n");
	printf("  -p number  select partition <number>\n");
	printf("Options for image/partition access:\n");
	printf("  -b         backup image/partition to ATR file\n");
	printf("             (this is the default)\n");
	printf("  -B         backup all images/partitions to files\n");
	printf("  -w         write ATR file to image/partition\n");
	printf("  -s number  only write first <number> sectors of partition to file\n");
	printf("             (default is 0/all sectors)\n");
	printf("  -fi        force writing of non standard image to image slot\n");
	printf("  -e num[-num] clear image information of given slots\n");
	printf("  @file      read parameters from file (one option/filename per line)\n");
	printf("  file...    ATR file(s) to read/write\n");
	printf("MyIDE Image Manager options:\n");
	printf("  -I         enable image manager support (implied by other -I opts)\n");
	printf("  -Ib        enable basic for images\n");
	printf("  -IB        disable basic for images (default)\n");
	printf("  -Ic num    set category (1..20)\n");
	printf("  -IC        list categories with name\n");
	printf("  -Ii        scan image slots and build new image manager table\n");
	printf("  -II        create empty image manager table\n");
	printf("  -IL        list image manager information\n");
	printf("  -Is[cCnNs] sort images by category number/name/slot number\n");
	printf("             upper-case selecs reverse sorting\n");
	printf("Moviewriter options:\n");
	printf("  -ma num    audio blocks per frame (default 3)\n");
	printf("  -mv num    video blocks per frame (default 27)\n");
	printf("  -m cylinder video-file audio-file\n");
	printf("             write video/audio files as movie\n");
	printf("  -mc cylinder video-file audio-file\n");
	printf("             write video/audio files as movie in compact flash mode\n");
	printf("             (512 bytes per sector)\n");

	return 1;
}

