#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "MyOpts.h"

// supported movie types
enum EMovieType {
	eGray64x216,
	eGray80x172,
	eColor128x64
};

inline unsigned int GetWidth(EMovieType type)
{
	switch(type) {
	case eGray64x216: return 64;
	case eGray80x172: return 80;
	case eColor128x64: return 128;
	}
	assert(false);
	return 0;
}

inline unsigned int GetHeight(EMovieType type)
{
	switch(type) {
	case eGray64x216: return 216;
	case eGray80x172: return 172;
	case eColor128x64: return 64;
	default: break;
	}
	assert(false);
	return 0;
}

// convert from 8bit to 4 bit
// return number of blocks converted

// read a pgm or ppm file and return raw image data if successful
// this function checks if the image size matches and also converts
// ppm (color) to pgm (grayscale) if necessary
unsigned char* read_ppm(EMovieType type, FILE* f, const char* name, bool color)
{
	char ppmtype;
	char tmpchar;
	unsigned char* data = 0;
	unsigned int width;
	unsigned int height;
	unsigned int maxval;
	unsigned int len;
	unsigned int len2;

	if (fscanf(f, "P%c %d %d %d%c", &ppmtype, &width, &height, &maxval, &tmpchar) != 5) {
		printf("cannot parse pgm header in %s\n", name);
		return 0;
	}
	if (color) {
		if (ppmtype != '6') {
			printf("illegal PPM format '%c' (expected '6')\n", ppmtype);
			return 0;
		}
	} else {
		if (ppmtype != '5') {
			printf("illegal PGM format '%c' (expected '5')\n", ppmtype);
			return 0;
		}
	}
	if ((width != GetWidth(type)) || (height != GetHeight(type))) {
		printf("illegal image dimensions of %s (%dx%d)\n", name, width, height);
		return 0;
	}
	if (maxval != 255) {
		printf("image %s is not in 8bit format (maxval = %d)\n", name, maxval);
		return 0;
	}

	len = width * height;
	if (color) {
		len = len * 3;
	}
	data = new unsigned char[len];

	if ((len2 = fread(data, 1, len, f)) != len) {
		printf("error reading %s : got %d of %d bytes\n", name, len2, len);
		delete[] data;
		return 0;
	}
	if (color && type != eColor128x64) {
		// convert to grayscale
		float fpix;
		int pix;
		unsigned int i;
		for (i=0; i<len/3; i++) {
			// standard CIE color->grayscale conversion 0.3*R + 0.59*G + 0.11*B
			fpix = 0.3 * data[i*3] + 0.59 * data[i*3+1] + 0.11 * data[i*3+2];
			pix = (int) (fpix + 0.5);
			//pix = (data[i*3] + data[i*3+1] + data[i*3+2]) / 3;
			if (pix < 0) {
				pix = 0;
			}
			if (pix > 255) {
				pix = 255;
			}
			data[i] = (unsigned char) pix;
		}
	}
	return data;
}

bool pack_6bit_color_frame(EMovieType type, const unsigned char* data, FILE* f)
{
	int rgb;
	unsigned int len = GetWidth(type)*GetHeight(type)/4;
	unsigned char buf[len];
	unsigned int i;
	unsigned char p1, p2, p3, p4;

	// 3 passes, first B then G then R
	for (rgb = 2; rgb >= 0; rgb--) {
		// input data is in RGB RGB RGB RGB ... format
		// combine 4 pixels into 1 byte
		for (i=0; i<len; i++) {
			p1 = data[i*4*3 + rgb] >> 6;
			p2 = data[i*4*3 + rgb + 3] >> 6;
			p3 = data[i*4*3 + rgb + 6] >> 6;
			p4 = data[i*4*3 + rgb + 9] >> 6;
			buf[i] = (p1 << 6) | (p2 << 4) | (p3 << 2) | p4;
		}
		if (fwrite(buf, 1, len, f) != len) {
			return false;
		}
	}
	return true;
}

bool pack_frame(EMovieType type, const unsigned char* data, FILE* f)
{
	if (type == eColor128x64) {
		return pack_6bit_color_frame(type, data, f);
	}

	// pack into 4bit grayscale images

	unsigned int len = GetWidth(type)*GetHeight(type)/2;
	unsigned char buf[len];
	unsigned char zero16[16];
	unsigned int i;
	memset(zero16, 0, 16);

	if (type == eGray80x172) {
		// this frame type needs 16 padding bytes at the beginning
		// and at the end of each frame
		if (fwrite(zero16, 1, 16, f) != 16) {
			return false;
		}

	}
	for (i=0;i<len;i++) {
		unsigned char p1 = data[i*2] & 0xf0;
		unsigned char p2 = data[i*2+1] >> 4;
		buf[i] = p1 | p2;
	}
	if (fwrite(buf, 1, len, f) != len) {
		return false;
	}
	if (type == eGray80x172) {
		// this frame type needs 16 padding bytes at the beginning
		// and at the end of each frame
		if (fwrite(zero16, 1, 16, f) != 16) {
			return false;
		}
	}
	return true;
}

// directory separator
#ifdef WIN32
#define DIR_SEPARATOR '\\'
#else
#define DIR_SEPARATOR '/'
#endif

int main(int argc, char** argv)
{
	unsigned long long frame = 0;
	char inname[PATH_MAX];
	bool colormode = false;

	const char* indir;
	const char* outname;
	const char* extension;
	FILE *outfile;

	EMovieType movieType;
       
	movieType = eGray64x216;
	// movieType = eGray80x172;

	unsigned long long step_in = 1; // 11988;
	unsigned long long step_out = 1; // 8000;

	printf("MyIDE movie converter V0.01 (c) 2007 by Matthias Reichl <hias@horus.com>\n");
	RCPtr<MyOpts> myopts = new MyOpts(argc, argv, true);

	const char* arg;
	while ((arg = myopts->GetOpt()) && arg[0] == '-') {
		switch (arg[1]) {
		case 't': // movie type
			if (!(arg = myopts->GetOpt())) {
				goto usage;
			}
			switch (atoi(arg)) {
			case 1:	
				movieType = eGray64x216;
				printf("Movie mode is 64x216 grayscale\n");
				break;
			case 2:
				movieType = eGray80x172;
				printf("Movie mode is 80x172 grayscale\n");
				break;
			case 3:
				movieType = eColor128x64;
				printf("Movie mode is 128x64 color\n");
				break;
			default:
				printf("illegal movie type \"%s\"\n", arg);
				goto usage;
			}
			continue;
		case 'p': // pulldown;
			if (!(arg = myopts->GetOpt())) {
				goto usage;
			}
			int in,out;
			if (sscanf(arg, "%d:%d", &in, &out) != 2) {
				printf("illegal pulldown option \"%s\"\n", arg);
				goto usage;
			}
			if (in <= 0) {
				printf("illegal input pulldown-value %d\n", in);
				goto usage;
			}
			if (out <= 0) {
				printf("illegal output pulldown-value %d\n", out);
				goto usage;
			}
			step_in = in;
			step_out = out;
			printf("Using a pulldown factor of %d:%d\n", (unsigned int) step_in, (unsigned int) step_out);
			continue;
		case 'c': // color mode
			colormode = true;
			break;
		case 'h':
			goto usage;
		default:
			printf("illegal option \"%s\"\n", arg);
			goto usage;
		}
	}

	if (!arg) {
		goto usage;
	}
	indir = arg;
	if (!(outname = myopts->GetOpt())) {
		goto usage;
	}

	if (movieType == eColor128x64) {
		colormode = true;
	}
	if (colormode) {
		printf("Color mode set: source pictures are in PPM format\n");
		extension = "ppm";
	} else {
		extension = "pgm";
	}

	outfile = fopen(outname, "wb");
	if (!outfile) {
		printf("error creating output file %s\n", outname);
		return 1;
	}

	frame = 0;
	while (1) {
		// calculate frame number based upon inout/output step
		unsigned long long frameno =
			(frame * step_in + (step_out >> 1) ) / step_out;
		//printf("frame %d -> %d\n", (unsigned int) frame, (unsigned int) frameno);
		sprintf(inname, "%s%c%08d.%s", indir, DIR_SEPARATOR, (unsigned int) frameno+1, extension);

		FILE* f = fopen(inname,"rb");
		if (!f) { // file not found
			if (frame == 0) {
				printf("cannot open first frame \"%s\", aborting\n", inname);
			}
			break;
		}

		unsigned char* data = read_ppm(movieType, f, inname, colormode);
		fclose(f);

		if (!data) {
			printf("aborting at frame %d\n", (unsigned int) frame+1);
			break;
		}
		if (!pack_frame(movieType, data, outfile)) {
			printf("writing frame %d failed\n", (unsigned int) frame+1);
			break;
		}
		delete[] data;
		frame++;
	}
	printf("wrote %d frames\n", (unsigned int) frame);

	fclose(outfile);
	return 0;

usage:
	printf("\nusage: myvideo [-t type] [-p <in>:<out>] imgdir outfile\n");
	printf("-t type        set movie type/image format:\n");
	printf("               1 = 64x216 gray (default)\n");
	printf("               2 = 80x172 gray\n");
	printf("               3 = 128x64 color\n");
	printf("-p <in>:<out>  set frame pulldown (default is 1:1)\n");
	printf("-c             color mode: read PPM (instead of PGM) pictures\n");
	printf("               note: this is automatically selected for color movies\n");
	printf("imgdir         directory containing movie images\n");
	printf("               image name must be an 8-digit framenumber starting at 1\n");
	printf("outfile        output filename\n");
	return 1;
}
